class SignatureError(Exception):
    """
    Raised when there are issues with signing.
    """


class ConfigurationError(Exception):
    """
    Raised when there are issues with the configuration.
    """
